#ifndef INCLUDE_GRAPHICS_IMPL_H
#define INCLUDE_GRAPHICS_IMPL_H

// graphics_impl.h
// Revision 18-feb-2006


#include "graphics.h"


#include "config_graphics.h"


// This is controlled with the configure option --disable-graphics
// If configure is not used, defining it here will disable the use
// of graphics.

#ifndef BLASSIC_CONFIG_NO_GRAPHICS


#include <queue>


namespace blassic {

namespace graphics_impl {


extern std::string program_name;
extern std::string default_title;


using blassic::graphics::Point;
using blassic::graphics::InkNum;
using blassic::graphics::RotateMode;
using blassic::graphics::RotateNone;
using blassic::graphics::Rotate90;
using blassic::graphics::TransformType;
using blassic::graphics::TrIdentity;
using blassic::graphics::TrInvertY;
using blassic::graphics::Inkset;
using blassic::graphics::InkStandard;
using blassic::graphics::InkCpc;
using blassic::graphics::InkSpectrum;
using blassic::graphics::ZoomValue;



// Define the same struct used in X
struct ShortPoint {
	short x, y;
};


template <class C>
inline void real_rotate (C & x, C & y, RotateMode mode, int screenwidth)
{
	switch (mode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		{
			C newx= y;
			y= static_cast <C> (screenwidth - x - 1);
			x= newx;
		}
		break;
	}
}

template <class C>
inline void real_unrotate (C & x, C & y, RotateMode mode, int screenwidth)
{
	switch (mode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		{
			C newy= x;
			x= static_cast <C> (screenwidth - y - 1);
			y= newy;
		}
		break;
	}
}

template <class C>
inline void real_rotate_rel (C & x, C & y, RotateMode mode)
{
	switch (mode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		{
			C newy= static_cast <C> (-x);
			x= y;
			y= newy;
		}
		break;
	}
}


template <class C>
inline void real_rotate_rel_abs (C & x, C & y, RotateMode mode)
{
	switch (mode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		{
			swap (x, y);
		}
		break;
	}
}

template <class C>
inline void sort_pair (C & n1, C & n2)
{
	if (n1 > n2)
		swap (n1, n2);
}

enum drawmodenums {
	// Numeric draw modes.
	// Used as index in mode tables and as values
	// used in programs to set it.
	drawmodenum_copy=   0,
	drawmodenum_xor=    1,
	drawmodenum_and=    2,
	drawmodenum_or=     3,
	// 0 to 3 are Amstrad CPC modes.
	drawmodenum_invert= 4,
	drawmodenum_max=    4
};


struct ColorRGB {
	int r;
	int g;
	int b;
	void set (int nr= 0, int ng= 0, int nb= 0)
	{
		r= nr; g= ng; b= nb;
	}
};

//enum Inkset { InkStandard, InkCpc, InkSpectrum };


//typedef unsigned int ZoomValue;

const ZoomValue MAXZOOMTEXTY= 32;
const ZoomValue MAXZOOMTEXTX= 32;

//const int MAXZOOMTEXT= std::max (MAXZOOMTEXTX, MAXZOOMTEXTY);
// Borland C++ can't evaluate this at compile time.
const ZoomValue MAXZOOMTEXT= 32;


const size_t MAXINKEYCODE= 79;


class QueueKey {
public:
	QueueKey ();
	virtual ~QueueKey ();
	void push (const std::string & str);
	std::string pop ();
	bool empty ();
	void erase ();
private:
	std::queue <std::string> q;
	virtual void lock ();
	virtual void unlock ();
};


class Image {
public:
	Image ();
	virtual ~Image ();
	virtual int getwidth () const = 0;
	virtual int getheight () const = 0;
};


class Painter {
public:
	Painter (int minx, int miny, int maxx, int maxy);
	virtual ~Painter ();
	void paint (int x, int y);
protected:
	virtual void plot (Point p)= 0;
	virtual bool do_check (Point p)= 0;
private:
	bool check (Point p);
	void gf (Point p);

	//typedef std::vector <char> v_visited_t;
	//typedef std::vector <v_visited_t> visited_t;
	//visited_t visited;
	class Visited {
	public:
		Visited (int width_n, int height_n);
		bool check (int x, int y) const;
		void set (int x, int y);
	private:
		std::vector <char> v;
		int width;
	};
	Visited visited;

	std::deque <Point> processlist, nextlist;
	int min_x, min_y, max_x, max_y;
};

class Screen;


class GraphicsSystem {
public:
	GraphicsSystem ();
	virtual ~GraphicsSystem ();
	virtual Screen * newScreen ()= 0;

	virtual void check_events ();
};



class Screen {
public:
	Screen (GraphicsSystem & system_n);
	virtual ~Screen ();

	GraphicsSystem & get_system ();

	bool is_active () const;

	void create (int width_n, int height_n, RotateMode rotmode,
		Inkset inkset_n, InkNum pen_n, InkNum paper_n,
		ZoomValue charx, ZoomValue chary);
	virtual void destroy ()= 0;

	virtual void flush ();

	virtual void update_window (int x, int y, int width, int height)= 0;
	virtual void update_window ()= 0;

	virtual void setink (InkNum inknum, const ColorRGB & rgb)= 0;
	void setcpcink (InkNum inknum, int cpccolor);
	void init_colors ();
	virtual void setactiveink (InkNum inknum)= 0;

	virtual void set_border (InkNum inknum)= 0;
	virtual void set_border_width (int newwidth)= 0;

	void set_graphics_pen (InkNum inknum);
	void set_graphics_paper (InkNum inknum);
	InkNum get_graphics_pen () const;
	InkNum get_graphics_paper () const;
	InkNum get_default_pen () const;
	InkNum get_default_paper () const;

	void set_transform (TransformType newtransform);

	virtual void set_title (const std::string & title)= 0;

	void cls ();

	void do_scroll (int x1, int y1, int x2, int y2, int xoff, int yoff,
		bool fill_it= false, InkNum fillink= InkNum (0) );
	void scroll (int x1, int y1, int x2, int y2, int xoff, int yoff,
		bool fill_it, InkNum fillink);

	void do_scroll_up (int x1, int y1, int x2, int y2,
		int amount, InkNum fillink);
	void do_scroll_down (int x1, int y1, int x2, int y2,
		int amount, InkNum fillink);
	void do_scroll_left (int x1, int y1, int x2, int y2,
		int amount, InkNum fillink);

	void set_origin (int x, int y);
	void set_limits (int minx, int maxx, int miny, int maxy);
	virtual void setdrawmode (int drawmodenum)= 0;
	void set_transparent (bool active);
	void set_mask (unsigned char newmask);
	void set_maskfirstpoint (bool firstpoint);

	int get_lastx () const { return lastx; }
	int get_lasty () const { return lasty; }
	void do_move (int x, int y);
	void do_move_rel (int x, int y);
	void plot (int x, int y);
	void plot_rel (int x, int y);
	int test (int x, int y, bool relative);
	void do_line (int x, int y);
	void do_line_rel (int x, int y);
	void do_drawarc (int x, int y, double g);
	void rectangle (Point org, Point dest);
	void do_fill_rectangle (int x1, int y1, int x2, int y2,
		bool limitable);
	void fill_rectangle (Point org, Point dest);
	void invert_rectangle (int x1, int y1, int x2, int y2);
	void zx_plot (Point p);
	void zx_unplot (Point p);
	void circle (int x, int y, int radius);
	void arccircle (int x, int y, int radius,
		BlNumber arcbeg, BlNumber arcend);
	void ellipse (int ox, int oy, int rx, int ry);
	void arcellipse (int ox, int oy, int rx, int ry,
		BlNumber arcbeg, BlNumber arcend);

	Image * getImage (int x1, int y1, int x2, int y2);
	void putImage (Image & image, int x, int y, int mode);

	void paint (InkNum inkpaint, InkNum inkborder, int x, int y);

	int get_maxtcol () const { return maxtcol; }
	int get_maxtrow () const { return maxtrow; }
	int getcharwidth () const { return charwidth; }
	int getcharheight () const { return charheight; }
	void print (unsigned char ch, int col, int row,
		InkNum chpen, InkNum chpaper,
		bool inverse, bool underline);
	void printxy (unsigned char ch,
		bool limitable, bool inverse, bool underline);

	virtual void set_synchro (bool synchro_n)= 0;
	virtual bool get_synchro () const= 0;
	void suspend_synchro ();
	void restart_synchro ();

	std::string copychrat (int x, int y, BlChar from, BlChar to);
	std::string copychratpos (int col, int row, BlChar from, BlChar to);

	//virtual void check_events ();
	virtual void wait_event ()= 0;

	virtual QueueKey & getQueueKey ()= 0;
	virtual int keypressed (int keynum)= 0;

	std::string inkey ();
	std::string getkey ();

	int get_xmouse () const;
	int get_ymouse () const;

	virtual void ring_my_bell () = 0;

protected:

	static const InkNum PREDEFINED_INKS= 16;

	virtual void show ()= 0;

	virtual int raw_xmouse () const= 0;
	virtual int raw_ymouse () const= 0;

private:
	GraphicsSystem & system;
	bool active;

	virtual void raw_create (int width, int height,
		int rwidth, int rheight,
		RotateMode rotmode,
		InkNum paperink)= 0;

	int get_width () const;
	int get_height () const;

	
	void set_standard_default_inks ();
	void set_cpc_default_inks ();
	void set_spectrum_default_inks ();

	virtual void raw_init_colors ()= 0;
	Inkset getInkset () const;

	bool check_limit (int x, int y);
	virtual void init_pixmap ()= 0;

	void do_line_unmasked (int x, int y);
	void impl_plot_mask (int x, int y);
	void do_line_mask (int x, int y);

	virtual void raw_copy_rectangle (int x1from, int y1from,
		int x2from, int y2from, int xto, int yto)= 0;

	virtual void raw_invert_rectangle (int x1, int y1, int x2, int y2)= 0;

	virtual void raw_rectangle (int x1, int y1, int x2, int y2)= 0;
	void do_plot (int x, int y);
	void do_drawarcsegment (int xd, int yd,
		int & coords_x, int & coords_y, bool & firstpoint);

	virtual void direct_fill_rectangle (int x1, int y1, int x2, int y2)= 0;
	virtual void raw_plot (int x, int y)= 0;
	void do_plot_points (ShortPoint point [], int npoints, bool limitable);
	virtual void real_plot_points (ShortPoint point [], int npoints)= 0;
	virtual void real_plot_wide_points (ShortPoint point [], int npoints,
		int widex, int widey,
		int xmin, int ymin, int xmax, int ymax)= 0;
	void zx (Point p);
	virtual int raw_test (int x, int y)= 0;

	virtual void raw_draw_line (int prevx, int prevy, int x, int y)= 0;
	virtual bool load_char_image (int x, int y, int xzoom, int yzoom,
		unsigned char (& ch) [8] )= 0;

	virtual Image * raw_getImage (int x1, int y1, int x2, int y2)= 0;
	virtual void raw_putImage (Image & image, int x, int y, int mode)= 0;

	virtual Painter * raw_getPainter (InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy)= 0;

	void do_printxy (unsigned char ch, int x, int y,
		InkNum chpen, InkNum chpaper,
		bool limitable, bool inverse, bool underline);

	Inkset inkset;
	int scwidth, scheight;
	int realwidth, realheight;
	ZoomValue zoomtextx;
	ZoomValue zoomtexty;
	int charwidth, charheight;
	int maxtcol, maxtrow;
	int lastx, lasty;

	TransformType activetransform;

	InkNum default_pen;
	InkNum default_paper;
	InkNum graphics_pen;
	InkNum graphics_paper;

	bool transparent_text;

	bool limited;
	int limit_minx, limit_miny, limit_maxx, limit_maxy;

	RotateMode rotatemode;

	size_t numsynchrosaved;
	bool synchrosaved;

	static const unsigned char maskvaluedefault= '\xFF';
	unsigned char maskvalue;
	bool maskdrawfirst;
	unsigned maskpos;
	static unsigned const char auxmask [];

	int originx;
	int originy;

	ZoomValue getzoomtextx () const { return zoomtextx; }

	void adjust_y (int & y) const;
	void transform_x (int & x) const;
	void transform_y (int & y) const;

	void do_rotate (int & x, int & y) const;
	void do_rotate (short int & x, short int & y) const;
	void do_rotate_rel (int & x, int & y) const;
	void do_rotate_rel (short int & x, short int & y) const;
	void do_rotate_rel_abs (int & x, int & y) const;
	void do_rotate_rel_abs (short int & x, short int & y) const;
	void do_unrotate (int & x, int & y) const;
	void do_unrotate (short int & x, short int & y) const;
};


extern GraphicsSystem * pcurrentsystem;


#ifdef BLASSIC_USE_WINDOWS_GRAPHICS
GraphicsSystem * createSystemMsWin ();
#endif

#ifdef BLASSIC_USE_X
std::string check_display (const std::string & sysname);
GraphicsSystem * createSystemXWin (const std::string & sysname);
#endif


} // namespace graphics_impl

} // namespace blassic

#endif


#endif

// End of graphics_impl.h
